@@ -16,6 +16,7 @@
 #elif __GLASGOW_HASKELL__ >= 702
 {-# LANGUAGE Trustworthy #-}
 #endif
+#include "bifunctors-common.h"
 
 -----------------------------------------------------------------------------
 -- |
@@ -53,6 +54,10 @@ import Data.Typeable
 import GHC.Generics
 #endif
 
+#if LIFTED_FUNCTOR_CLASSES
+import Data.Functor.Classes
+#endif
+
 -- | Make a 'Functor' over the second argument of a 'Bifunctor'.
 newtype WrappedBifunctor p a b = WrapBifunctor { unwrapBifunctor :: p a b }
   deriving ( Eq, Ord, Show, Read
@@ -89,6 +94,37 @@ instance Generic1 (WrappedBifunctor p a) where
     to1 = WrapBifunctor . unRec1 . unM1 . unM1 . unM1
 #endif
 
+#if LIFTED_FUNCTOR_CLASSES
+instance (Eq2 p, Eq a) => Eq1 (WrappedBifunctor p a) where
+  liftEq = liftEq2 (==)
+instance Eq2 p => Eq2 (WrappedBifunctor p) where
+  liftEq2 f g (WrapBifunctor x) (WrapBifunctor y) = liftEq2 f g x y
+
+instance (Ord2 p, Ord a) => Ord1 (WrappedBifunctor p a) where
+  liftCompare = liftCompare2 compare
+instance Ord2 p => Ord2 (WrappedBifunctor p) where
+  liftCompare2 f g (WrapBifunctor x) (WrapBifunctor y) = liftCompare2 f g x y
+
+instance (Read2 p, Read a) => Read1 (WrappedBifunctor p a) where
+  liftReadsPrec = liftReadsPrec2 readsPrec readList
+instance Read2 p => Read2 (WrappedBifunctor p) where
+  liftReadsPrec2 rp1 rl1 rp2 rl2 p = readParen (p > 10) $ \s0 -> do
+    ("WrapBifunctor",   s1) <- lex s0
+    ("{",               s2) <- lex s1
+    ("unwrapBifunctor", s3) <- lex s2
+    (x,                 s4) <- liftReadsPrec2 rp1 rl1 rp2 rl2 0 s3
+    ("}",               s5) <- lex s4
+    return (WrapBifunctor x, s5)
+
+instance (Show2 p, Show a) => Show1 (WrappedBifunctor p a) where
+  liftShowsPrec = liftShowsPrec2 showsPrec showList
+instance Show2 p => Show2 (WrappedBifunctor p) where
+  liftShowsPrec2 sp1 sl1 sp2 sl2 p (WrapBifunctor x) = showParen (p > 10) $
+      showString "WrapBifunctor {unwrapBifunctor = "
+    . liftShowsPrec2 sp1 sl1 sp2 sl2 0 x
+    . showChar '}'
+#endif
+
 instance Bifunctor p => Bifunctor (WrappedBifunctor p) where
   first f = WrapBifunctor . first f . unwrapBifunctor
   {-# INLINE first #-}
